'''
Created on 2023-07-19

@author: wf
'''
from nicescad.version import Version
import sys
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
import traceback
import webbrowser
from nicescad.webserver import WebServer

def getArgParser(description:str,version_msg)->ArgumentParser:
    """
    Setup command line argument parser
    
    Args:
        description(str): the description
        version_msg(str): the version message
        
    Returns:
        ArgumentParser: the argument parser
    """
    parser = ArgumentParser(description=description, formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument("-a","--about",help="show about info [default: %(default)s]",action="store_true")
    parser.add_argument("-c","--client", action="store_true", help="start client [default: %(default)s]")
    parser.add_argument("-d", "--debug", dest="debug", action="store_true", help="show debug info [default: %(default)s]")
    parser.add_argument("-i", "--input", help="input file")

    parser.add_argument("--host", default="localhost",
                            help="the host to serve / listen from [default: %(default)s]")
    parser.add_argument("--port",type=int,default=9858,help="the port to serve from [default: %(default)s]")
    parser.add_argument("-s","--serve", action="store_true", help="start webserver [default: %(default)s]")
    parser.add_argument("-V", "--version", action='version', version=version_msg)
    return parser


def main(argv=None): 
    '''main program.'''

    if argv is None:
        argv=sys.argv[1:]
        
    program_name = Version.name
    program_version =f"v{Version.version}" 
    program_build_date = str(Version.date)
    program_version_message = f'{program_name} ({program_version},{program_build_date})'

    try:
        parser=getArgParser(description=Version.license,version_msg=program_version_message)
        args = parser.parse_args(argv)
        if len(argv) < 1:
            parser.print_usage()
            sys.exit(1)
        if args.about:
            print(program_version_message)
            print(f"see {Version.doc_url}")
            webbrowser.open(Version.doc_url)
        if args.client:
            url=f"http://{args.host}:{args.port}"
            webbrowser.open(url)
        if args.serve:
            ws=WebServer()
            ws.run(args)
        
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 1
    except Exception as e:
        if DEBUG:
            raise(e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        if args.debug:
            print(traceback.format_exc())
        return 2       
        
DEBUG = 1
if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(main())


"""
Created on 2023-06-19

@author: wf
"""
from typing import List, Optional
from nicescad.version import Version
from nicescad.openscad import OpenScad
from nicescad.local_filepicker import LocalFilePicker
from nicegui import ui, app
import os
import sys
import requests
import traceback

class WebServer:
    """WebServer class that manages the server and handles OpenScad operations.

    Attributes:
        oscad (OpenScad): An OpenScad object that aids in performing OpenScad operations.
    """

    def __init__(self):
        """Constructs all the necessary attributes for the WebServer object."""
        self.oscad = OpenScad()
        self.code="""c=cube(3;center=true);
s=sphere(2,center=true);"""        
        self.input="example.stl"
        self.is_local=True
        app.add_static_files('/stl', self.oscad.tmp_dir)
        self.error_area=None
        self.do_trace=True

        @ui.page('/')
        def home():
            self.home()
            
        @ui.page('/settings')
        def settings():
            self.settings()
            
            
    def handle_exception(self, e: Exception, trace: Optional[bool] = False):
        """Handles an exception by creating an error message.

        Args:
            e (Exception): The exception to handle.
            trace (bool, optional): Whether to include the traceback in the error message. Default is False.
        """
        if trace:
            self.error_msg = str(e) + "\n" + traceback.format_exc()
        else:
            self.error_msg = str(e)
        if self.error_area:
            self.error_area.text=self.error_msg
        print(self.error_msg,file=sys.stderr)

 
    def render(self, click_args):
        """Renders the OpenScad string and updates the 3D scene with the result.

        Args:
            click_args (object): The click event arguments.
        """
        openscad_str = self.code_area.value
        stl = self.oscad.openscad_str_to_file(openscad_str)
        with self.scene:
            self.scene.clear()
            self.scene.stl("/stl/tmp.stl").move(x=-0.5).scale(0.06)
            
    def do_read_input(self, input: str):
        """Reads the given input.

        Args:
            input (str): The input string representing a URL or local path.
        """
        if input.startswith('http://') or input.startswith('https://'):
            response = requests.get(input)
            if response.status_code == 200:
                return response.text
            else:
                raise Exception(f'Unable to retrieve data from URL: {input}')
        else:
            if os.path.exists(input):
                with open(input, 'r') as file:
                    return file.read()
            else:
                raise Exception(f'File does not exist: {input}')
    
    def read_input(self, input: str):
        """Reads the given input and handles any exceptions.

        Args:
            input (str): The input string representing a URL or local file.
        """
        self.input=input
        try:
            self.code = self.do_read_input(input)
            self.error_msg = None
        except Exception as e:
            self.code = None
            self.handle_exception(e, self.do_trace)
            
    def save_file(self):
        """Saves the current code to the last input file, if it was a local path."""
        if self.is_local and self.input:
            with open(self.input, 'w') as file:
                file.write(self.code)
        else:
            raise Exception('No local file to save to')
    
    async def open_file(self) -> None:
        """Opens a dialog."""
        result = await LocalFilePicker('~', multiple=True)
        ui.notify(f'You chose {result}')
    pass
    
    def menu(self):
        """Adds a link to the project's GitHub page in the web server's menu."""
        ui.link('nicescad on GitHub', 'https://github.com/WolfgangFahl/nicescad')
        with ui.row().classes('w-full items-center'):
            result = ui.label().classes('mr-auto')
            with ui.button(icon='menu'):
                with ui.menu() as menu:
                    ui.menu_item('Open', self.open_file)
                    ui.menu_item('Save', self.save_file)
                    ui.separator()
                    ui.menu_item('Close', on_click=menu.close)
       
    def home(self):
        """Generates the home page with a 3D viewer and a code editor."""
        self.menu()
        with ui.splitter() as splitter:
            with splitter.before:
                with ui.scene(width=1024, height=768) as scene:
                    self.scene = scene
                    scene.spot_light(distance=100, intensity=0.1).move(-10, 0, 10)
            with splitter.after:
                self.input_input=ui.input(value=self.input)
                self.code_area = ui.textarea(value=self.code).props('clearable;rows=20')
                self.error_area = ui.textarea()
                ui.button('Render', on_click=self.render)
        if self.args.input:
            self.read_input(self.args.input)
  
        
    def settings(self):
        """Generates the settings page with a link to the project's GitHub page."""
        self.menu()
       
    def run(self, args):
        """Runs the UI of the web server.

        Args:
            args (list): The command line arguments.
        """
        self.args=args
        ui.run(title=Version.name, host=args.host, port=args.port, reload=False)



'''
Created on 2023-07-19

@author: wf
'''
from tests.basetest import Basetest
from nicescad.openscad import OpenScad

class TestOpenScad(Basetest):
    """
    test openscad wrapper
    """
    
    def test_detect_openscad_exec(self):
        """
        test finding the openscad executable
        """
        oscad=OpenScad()
        self.assertTrue(oscad.openscad_exec is not None)
        
    def testRender(self):
        """
        test rendering a scad file
        """
        # openscad_exec="/Users/wf/bin/openscad"
        oscad=OpenScad()
        openscad_str="cube(5);"
        stl=oscad.openscad_str_to_file(openscad_str)
        print(stl)
        pass
        


"""
Created on 2023-07-19

@author: wf

This module contains the class OpenScad, a wrapper for OpenScad.
"""

from typing import Dict
import tempfile
import os
import platform
import subprocess


class OpenScad:
    """
    A wrapper for OpenScad (https://openscad.org/).
    """

    def __init__(self,**kw) -> None:
        """
        Initializes the OpenScad object.
        """
        self.openscad_exec = None
        self.openscad_tmp_dir = None
        if 'OPENSCAD_EXEC' in os.environ: self.openscad_exec = os.environ['OPENSCAD_EXEC']
        if 'OPENSCAD_TMP_DIR' in os.environ: self.openscad_tmp_dir = os.environ['OPENSCAD_TMP_DIR']
        if self.openscad_tmp_dir is not None:
            self.tmp_dir = self.openscad_tmp_dir
        else:
            self.tmp_dir = tempfile.mkdtemp()
        if 'openscad_exec' in kw: self.openscad_exec = kw['openscad_exec']
        if self.openscad_exec is None:
            self._try_detect_openscad_exec()
        if self.openscad_exec is None:
            raise Exception('openscad exec not found!')
     
    def _try_executable(self, executable_path: str) -> None:
        """
        Checks if the specified path is a file. If it is, sets it as the OpenScad executable.

        Args:
            executable_path (str): The path to the executable file.
        """
        if os.path.isfile(executable_path):
            self.openscad_exec = executable_path
               
    def _try_detect_openscad_exec(self) -> None:
        """
        Tries to find the OpenScad executable on the system.

        References:
            https://github.com/nickc92/ViewSCAD/blob/d4597ff6870316dfaafa4f9ecc8ef62773081c61/viewscad/renderer.py#L206C5-L222C1
        """
        platfm = platform.system()
        if platfm == 'Linux':
            self._try_executable('/usr/bin/openscad')
            if self.openscad_exec is None:
                self._try_executable('/usr/local/bin/openscad')
        elif platfm == 'Darwin':
            self._try_executable('/Applications/OpenSCAD.app/Contents/MacOS/OpenSCAD')
        elif platfm == 'Windows':
            self._try_executable(os.path.join(os.environ.get('Programfiles(x86)', 'C:'), 'OpenSCAD\\openscad.exe'))
            self._try_executable(os.path.join(os.environ.get('Programfiles', 'C:'), 'OpenSCAD\\openscad.exe'))

    def render_to_file(self, openscad_str: str, fl_name: str, **kwargs) -> None:
        """
        Renders an OpenSCAD string to a file.

        Args:
            openscad_str (str): The OpenSCAD code.
            fl_name (str): The name of the output file.
            kwargs: Additional arguments, could be 'dollar_sign_vars' for dollar-sign variables or 'rough' for rough rendering.
        
        Raises:
            Exception: If there's an error in the OpenSCAD command.
        """
        scad_prepend = ''        
        dollar_sign_vars: Dict[str, str] = kwargs.get('dollar_sign_vars', {})
        for var_name, value in dollar_sign_vars.items():
            scad_prepend += '${}={};\n'.format(var_name, value)

        if not kwargs.get('rough', False) and not dollar_sign_vars:
            scad_prepend += '$fn=120;\n'
                
        scad_tmp_file = os.path.join(self.tmp_dir, 'tmp.scad')
        try:
            with open(scad_tmp_file, 'w') as of:
                of.write(scad_prepend)
                of.write(openscad_str)

            # now run openscad to generate stl:
            cmd = [self.openscad_exec, '-o', fl_name, scad_tmp_file]
            out = subprocess.check_output(cmd)
            if out != b'':
                print(out)
        except Exception as e:
            raise e
        finally:
            if os.path.isfile(scad_tmp_file):
                os.remove(scad_tmp_file) 
                
    def openscad_str_to_file(self, openscad_str, **kw):
        """
        render the openscad code to a file
        """
        self.saved_umask = os.umask(0o077)        
        if 'outfile' in kw:
            openscad_out_file = kw['outfile']
        else:
            openscad_out_file = os.path.join(self.tmp_dir, 'tmp.stl')                    
        try:
            self.render_to_file(openscad_str, openscad_out_file, **kw)
            if openscad_out_file.find('.stl') >= 0:                
                return openscad_out_file
            else:
                # @TODO improve
                print('No rendering if non-STL file is being created.')
        except Exception as e:
            raise e
        return None
